import React, { createContext, useContext, useMemo, useLayoutEffect, forwardRef } from 'react';
import classNames from 'classnames';
import shortId from 'shortid';
import stylis from 'stylis';
import invariant from 'invariant';
import { getContrast, readableColor } from 'polished';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
}

// @pragma export
function css(strings, ...values) {
    let combined = '';
    for (let i = 0; i < strings.length; i += 1) {
        const currentString = strings[i];
        const currentValue = values[i] || '';
        combined += currentString + currentValue;
    }
    return combined;
}

var ThemeContext = createContext(null);

// @pragma export
function useTheme() {
    const theme = useContext(ThemeContext);
    invariant(theme, '[hacker-ui] could not find theme context. ensure your app is wrapped with a theme provider.');
    return theme;
}

// @pragma export
const hasBadSurfaceContrast = (a, b) => getContrast(a, b) < 1.5;
const hasUnreadableContrast = (a, b) => getContrast(a, b) < 3;
function createDynamicColorPalette(color, onColor) {
    const asBackground = hasBadSurfaceContrast(color, onColor)
        ? readableColor(onColor)
        : color;
    const bgContrast = readableColor(asBackground);
    const onSurface = hasUnreadableContrast(color, onColor)
        ? readableColor(onColor)
        : color;
    return {
        asBackground,
        bgContrast,
        onSurface,
    };
}

function tryGetCurrentFileName() {
    if (process.env.NODE_ENV === 'production')
        return '';
    // TODO: try this out in other browsers
    const e = new Error();
    const { stack } = e;
    if (!stack)
        return '';
    const match = /at.*\/(.*)\.(?:t|j)sx?.*\(.*\)/.exec(stack);
    if (!match)
        return '';
    const fileName = match[1];
    return fileName.replace(/\./g, '-').replace(/[^a-z-_]/gi, '-');
}

function hashStyleObj(styleObj) {
    return Object.keys(styleObj)
        .map(key => `${key}_${styleObj[key]}`)
        .join('__|__');
}
// preserve the object reference
const empty = {};
function makeStyles(stylesFn) {
    const sheetId = shortId();
    const fileName = tryGetCurrentFileName();
    const sheetEl = document.createElement('style');
    sheetEl.dataset.hackerUi = 'true';
    sheetEl.id = sheetId;
    document.head.appendChild(sheetEl);
    function useStyles(props, component) {
        const theme = useTheme();
        const { color = theme.colors.bland, on = theme.colors.surface, style: incomingStyle, className: incomingClassName, styles: incomingStyles = empty } = props, restOfProps = __rest(props, ["color", "on", "style", "className", "styles"]);
        const incomingStyleHash = hashStyleObj(incomingStyles);
        // create a map of unprocessed styles
        const unprocessedStyles = useMemo(() => {
            const colors = createDynamicColorPalette(color, on);
            return stylesFn(css, colors);
        }, [color, on]);
        // calculate the class names
        const thisStyles = useMemo(() => {
            return Object.keys(unprocessedStyles)
                .map(key => [
                key,
                // the replace is ensure the class name only uses css safe characters
                `${fileName || 'hui'}_${key}_${sheetId}`.replace(/[^a-z0-9-_]/gi, ''),
            ])
                .reduce((acc, [key, className]) => {
                acc[key] = className;
                return acc;
            }, {});
        }, [unprocessedStyles]);
        // mount the styles to the dom
        useLayoutEffect(() => {
            const keys = Object.keys(thisStyles);
            const processedSheet = keys
                .map(key => {
                const className = thisStyles[key];
                const unprocessedStyle = unprocessedStyles[key];
                const processedStyle = stylis(`.${className}`, unprocessedStyle);
                return processedStyle;
            })
                .join('\n\n');
            sheetEl.innerHTML = processedSheet;
        }, [thisStyles, unprocessedStyles]);
        const mergedStyles = useMemo(() => {
            const thisStyleKeys = Object.keys(thisStyles);
            return thisStyleKeys.reduce((merged, key) => {
                const thisStyle = thisStyles[key];
                const incomingStyle = incomingStyles[key];
                merged[key] = classNames(thisStyle, incomingStyle);
                return merged;
            }, {});
            // eslint-disable-next-line react-hooks/exhaustive-deps
        }, [thisStyles, incomingStyleHash]);
        const Component = (component || 'div');
        const Root = useMemo(() => {
            return forwardRef((rootProps, ref) => {
                const { className: rootClassName, style: rootStyles } = rootProps;
                return (React.createElement(Component, Object.assign({}, rootProps, { ref: ref, className: classNames(mergedStyles.root, rootClassName, incomingClassName), style: Object.assign(Object.assign({}, rootStyles), incomingStyle) })));
            });
            // eslint-disable-next-line react-hooks/exhaustive-deps
        }, [incomingClassName, incomingStyleHash, mergedStyles.root]);
        return Object.assign({ Root, styles: mergedStyles }, restOfProps);
    }
    return useStyles;
}

const useStyles = makeStyles(css => ({
    root: css `
    background-color: blue;
  `,
}));
const Button = forwardRef((props, ref) => {
    const _a = useStyles(props, 'button'), { Root, styles } = _a, restOfProps = __rest(_a, ["Root", "styles"]);
    return React.createElement(Root, Object.assign({}, restOfProps));
});

// @pragma export
function ThemeProvider({ theme, children }) {
    return (React.createElement(ThemeContext.Provider, { value: theme }, children));
}

const defaultTheme = {
    colors: {
        brand: '#000',
        accent: '#2962ff',
        bland: '#ddd',
        danger: '#eb002b',
        warning: '#f56200',
        info: '#2962ff',
        surface: '#fff',
    },
};

function isObject(maybe) {
    if (!maybe)
        return false;
    return typeof maybe === 'object';
}
/**
 * recursively merges a base object with an incoming object
 */
function recursiveMerge(base, incoming = {}) {
    const keys = [...Object.keys(base), ...Object.keys(incoming)];
    const merged = keys.reduce((merged, key) => {
        const baseValue = base[key];
        const incomingValue = incoming[key];
        const mergedValue = isObject(baseValue)
            ? recursiveMerge(baseValue, incomingValue)
            : (incomingValue || baseValue);
        merged[key] = mergedValue;
        return merged;
    }, {});
    return merged;
}
function createTheme(partialTheme = {}) {
    return recursiveMerge(defaultTheme, partialTheme);
}

export { Button, ThemeContext, ThemeProvider, createDynamicColorPalette, makeStyles as createStyles, createTheme, css, defaultTheme, useTheme };
